from typing import Optional
from pydantic import BaseModel


class Grade(BaseModel):
    grade_abbrev: str
    grade_name: str
    grade_desc: str
    grade_min: int
    grade_max: int

    def __hash__(self):
        return hash((self.grade_abbrev))


from typing import Optional
import uuid
from pydantic import BaseModel


class AssignmentCriteria(BaseModel):
    criteria_id: str
    assignment_id: str
    criteria_desc: str
    criteria_weight: float

    def __hash__(self):
        return hash((self.criteria_desc, self.assignment_id))



from dataclasses import dataclass

@dataclass
class PgAssignmentStudentScoresTable:
    name: str = "assignment_student_scores"
    assignment_student_score: str = 'assignment_student_score'
    assignment_grade_abbrev: str = 'assignment_grade_abbrev'
    assignment_grade_desc: str = 'assignment_grade_desc'
    assignment_student_score_comments: str = 'assignment_student_score_comments'



from typing import Optional
from pydantic import BaseModel

class Assignment(BaseModel):
    assignment_id: str
    assignment_name: str
    assignment_abbrev: str
    session_id: str
    canvas_assignment_id: int

    def __hash__(self):
        return hash((self.assignment_abbrev, self.session_id, self.canvas_assignment_id))



import os
import pickle 
import json
from sqlalchemy import create_engine, MetaData, Table, and_
from sqlalchemy.sql import column
from pathlib import Path
from pydantic import BaseModel

from utsam.bases.params import _METADATA_


class PgDatabase:
    def __init__(
        self,
        user_name: str = None,
        password: str = None,
        host: str = None,
        db_name: str = None,
        port: str = '5432',
        echo = False,
        engine = None,
    ):
        self.database_type = 'postgresql'
        self.user_name = user_name
        self.password = password
        self.host = host
        self.db_name = db_name
        self.port = port
        self.connection_str = f'{self.database_type}://{self.user_name}:{self.password}@{self.host}:{self.port}/{self.db_name}'
        self.engine = engine
        self.create_engine()

    def execute_sql(self, sql_statement):
        try:
            #print(sql_statement)
            with self.engine.connect() as conn:
                if sql_statement.is_select:
                    return conn.execute(sql_statement).fetchall()
                conn.execute(sql_statement)
                conn.commit()
        except Exception as e:
            print(f"Error when writing to database: {e}")

    def create_engine(self):
        if self.engine is None:
            self.engine = create_engine(self.connection_string, pool_size=10, max_overflow=0, echo=False)


class PgTable(PgDatabase):
    def __init__(
        self,
        table_name: str = None,
        user_name: str = None,
        password: str = None,
        host: str = None,
        db_name: str = None,
        port: str = '5432',
        schema: str = "public",
        engine = None
    ):
        super().__init__(
            user_name = user_name,
            password = password,
            host = host,
            db_name = db_name,
            port = port,
            engine = engine,
        )
        self.table_name= table_name
        self.schema = schema
        self.metadata = None
        self.table = None
        self.meta_path = Path.cwd() / _METADATA_
        self.meta_file_name = f'{self.table_name}.metadata'
        self.load_metadata()

    def load_metadata(self):
        if os.path.isfile(self.meta_path / self.meta_file_name):  
            self.reload_metadata()
        else:
            self.metadata = MetaData(schema=self.schema)
            self.metadata.reflect(self.engine, only=[self.table_name])
            self.table = Table(self.table_name, self.metadata, autoload=True, autoload_with=self.engine)
            self.meta_path.mkdir(parents=True, exist_ok=True)

            temp_table = self.table.to_metadata(MetaData())
            with open(self.meta_path / self.meta_file_name, 'wb') as f:
                pickle.dump(temp_table, f)

    def reload_metadata(self):
        with open(self.meta_path / self.meta_file_name, 'rb') as f:
            self.table = pickle.load(f)
    
    def set_select_statement(self, where_dict: dict = None):
        statement = self.table.select()
        if where_dict:
            where_filters = [column(key) == value for key, value in where_dict.items()]
            statement = statement.where(and_(*where_filters))
        return statement

    def select_table(self, where_dict: dict = None):
        sql_statement = self.set_select_statement(where_dict=where_dict)
        return self.execute_sql(sql_statement=sql_statement)

    def set_insert_statement(self, records: list[BaseModel]):
        return (
            self.table.insert()
            .values([record.model_dump() for record in records])
        )

    def insert_records(self, records: list[BaseModel]):
        sql_statement = self.set_insert_statement(records=records)
        self.execute_sql(sql_statement=sql_statement)

    def set_update_statement(self, records: list[BaseModel], where_dict: dict = None):
        statement = self.table.update()
        if where_dict:
            where_filters = [column(key) == value for key, value in where_dict.items()]
            statement = statement.where(and_(*where_filters))
        return (
            statement
            .values([record.model_dump() for record in records])
        )

    def update_records(self, records: list[BaseModel], where_dict: dict = None):
        sql_statement = self.set_update_statement(records=records, where_dict=where_dict)
        self.execute_sql(sql_statement=sql_statement)

    #def set_upsert_statement(self, records: list[BaseModel], where_dict: dict = None):
    #    if len(self.select_table(where_dict=where_dict)) > 0:
    #        self.sql_statement = self.set_update_statement(records=records, where_dict=where_dict)
    #    else:
    #        self.sql_statement = self.set_insert_statement(records=records)
#
    #def upsert_records(self, records: list[BaseModel], where_dict: dict = None):
    #    sql_statement = self.set_upsert_statement(records=records, where_dict=where_dict)
    #    self.execute_sql(sql_statement=sql_statement)

def records_to_json(records):
    return json.loads(json.dumps([dict(row._mapping) for row in records], default=str))

def select_table(table_name, engine, where_dict=None, json=True):

    pg_table = PgTable(
        table_name=table_name,
        engine=engine,
    )

    records = pg_table.select_table(where_dict)
    if json:
        return records_to_json(records)
    return records
    

def insert_records(table_name, records, engine):

    pg_table = PgTable(
        table_name=table_name,
        engine=engine,
    )
    previous_len = len(pg_table.select_table())

    pg_table.insert_records(records=records)
    assert len(pg_table.select_table()) == previous_len + len(records)
    


from typing import Optional
from pydantic import BaseModel


class AssignmentGroup(BaseModel):
    group_id: str
    group_name: str
    assignment_id: str
    employee_id: str

    def __hash__(self):
        return hash((self.group_name, self.assignment_id, self.employee_id))



from dataclasses import dataclass

@dataclass
class PgAssignmentSubcriteriaGrade:
    name: str = "assignment_sassignment_subcriteria_gradesubcriterias"
    subcriteria_grade_id: str = 'subcriteria_grade_id'
    subcriteria_grade_desc: str = 'subcriteria_grade_desc'



from typing import Optional
from pydantic import BaseModel


class StudentGroup(BaseModel):
    group_id: str
    student_id: int
    canvas_group_id: Optional[int] = None

    def __hash__(self):
        return hash((self.student_id, self.canvas_group_id))



from typing import Optional
import uuid
from pydantic import BaseModel


class AssignmentSubCriteria(BaseModel):
    subcriteria_id: str
    criteria_id: str
    subcriteria_name: str
    subcriteria_desc: str
    subcriteria_weight: float

    def __hash__(self):
        return hash((self.criteria_id, self.subcriteria_name))



from typing import Optional
from pydantic import BaseModel


class Session(BaseModel):
    session_id: str
    session_year: int
    course_id: str
    semester_id: str
    
    def __hash__(self):
        return hash((self.session_year, self.course_id, self.semester_id))


from dataclasses import dataclass

@dataclass
class PgCoursesTable:
    name: str = "courses"
    course_id: str = 'course_id'
    course_code: str = 'course_code'
    course_name: str = 'course_name'
    course_abbrev: str = 'course_abbrev'


from sqlalchemy import create_engine
from pathlib import Path

from utsam.bases.params import SemesterParams, _CACHE_


def create_pg_engine(
    database_type: str,
    username: str,
    password: str,
    host: str,
    port: str,
    database_name: str,
    echo = False
    ):
    # Create a connection string
    connection_string = f'{database_type}://{username}:{password}@{host}:{port}/{database_name}'

    # Create a database engine
    return create_engine(connection_string, pool_size=10, max_overflow=0, echo=echo)

import os
import pandas as pd


def load_data(table_name, engine, load_cache=True, col_name='', filter_value=''):
    Path(_CACHE_).mkdir(parents=True, exist_ok=True)
    file_path = f"{_CACHE_}/{table_name}.csv"
    if not load_cache and os.path.isfile(file_path):
        print(f"loading cache from {file_path}")
        return pd.read_csv(file_path)

    filter_condition = ''
    if col_name and filter_value:
        filter_condition = f"WHERE {col_name} = '{filter_value}'"

    df = pd.read_sql(
        f"SELECT * FROM {table_name} {filter_condition};",
        engine
    )
    print(f"writing cache to {file_path}")
    df.to_csv(file_path, index=False)
    return df



from typing import Optional
from pydantic import BaseModel


class Course(BaseModel):
    course_id: str
    course_code: int
    course_name: str
    course_abbrev: str

    def __hash__(self):
        return hash((self.course_code))


import time
import json
from pathlib import Path
import pandas as pd
from sqlalchemy import update


from utsam.bases.params import _SELECT_, _LOGS_
from utsam.dbs.schemas.masterdata.employees import PgEmployeesTable
from utsam.dbs.schemas.masterdata.students import PgStudentsTable
from utsam.dbs.schemas.masterdata.assignments import PgAssignmentsTable
from utsam.dbs.schemas.masterdata.criterias import PgAssignmentCriteriasTable
from utsam.dbs.schemas.masterdata.subcriterias import PgAssignmentSubcriteriasTable
from utsam.dbs.schemas.transactions.assignment_criteria_student_scores import PgAssignmentCriteriaStudentScoresTable
from utsam.dbs.schemas.transactions.assignment_subcriteria_student_scores import PgAssignmentSubcriteriaStudentScoresTable

from utsam.bases.grades import get_grade


def get_groups_by_employee(employee, data):
    employees_mask = data[PgEmployeesTable.employee_first_name] == employee
    temp_df = data.loc[employees_mask, ]
    if len(temp_df) > 0:
        return [_SELECT_] +  list(set(temp_df[PgStudentsTable.group_name].sort_values().unique()))
    return [_SELECT_]

def get_students_ids_by_employee_and_group(employee, group_name, data):
    employees_mask = data[PgEmployeesTable.employee_first_name] == employee
    group_mask = data[PgStudentsTable.group_name] == group_name
    temp_df = data.loc[employees_mask & group_mask, ]
    if len(temp_df) > 0:
        return [_SELECT_] +  list(set(temp_df[PgStudentsTable.student_id].sort_values().unique()))
    return [_SELECT_]

def get_students_ids_by_employee(employee, data):
    employees_mask = data[PgEmployeesTable.employee_first_name] == employee
    temp_df = data.loc[employees_mask, ]
    if len(temp_df) > 0:
        return [_SELECT_] +  list(set(temp_df[PgStudentsTable.student_id].sort_values().unique()))
    return [_SELECT_]
    
def get_student_name(student_id, data):
    if student_id == _SELECT_:
        return None
    student_mask = data[PgStudentsTable.student_id] == int(student_id)
    temp_df = data.loc[student_mask, ]
    student_name = temp_df[PgStudentsTable.student_name].values
    if len(student_name) > 0:
        return student_name[0]
    return None


class StudentMarking:
    def __init__(
        self,
        student_id,
        student_name,
        tutor_name,
        assignment,
        grades_df: pd.DataFrame,
        engine,
    ):
        self.student_id = student_id
        self.student_name = student_name
        self.tutor_name = tutor_name
        self.assignment = assignment
        self.grades_df = grades_df.copy()
        self.criteria_dict: dict = {}
        self.score = 0
        self.grade = None
        self.engine = engine

    def print(self):
        print(f"Final Mark for '{self.student_name}' ({self.student_id}) = {self.score} ({self.grade})")
    
    def reset_score_and_grade(self):
        self.score = 0
        self.grade = None

    def append_criteria(self, criteria):
        self.criteria_dict[criteria.criteria_id] = criteria

    def set_score_and_grade(self):
        if len(self.criteria_dict) > 0:
            self.reset_score_and_grade()
            for criteria_id, crit in self.criteria_dict.items():
                print(f"{criteria_id} -> crit.score: {crit.score} ({type(crit.score)})")
                print(f"{criteria_id} -> crit.weight: {crit.weight} ({type(crit.weight)})")
            self.score = sum([crit.score * crit.weight for _, crit in self.criteria_dict.items()])
            self.grade = get_grade(self.grades_df, self.score)

    def to_dict(self):
        if len(self.criteria_dict) > 0:
            return {
                PgStudentsTable.student_id: self.student_id,
                PgStudentsTable.student_name: self.student_name,
                "tutor_name": self.tutor_name,
                "course_code": self.assignment.course_code,
                "semester": self.assignment.semester,
                "year": self.assignment.year,
                "assignment": self.assignment.assignment,
                "grade": self.grade,
                "score": self.score,
                #"comments": self.comments,
                "criterias": {crit_id: crit.to_dict() for crit_id, crit in self.criteria_dict.items()}
            }

    def to_json(self):
        logs_paths = Path.cwd() / _LOGS_ / self.tutor_name
        logs_paths.mkdir(parents=True, exist_ok=True)
        json_data = self.to_dict()
        json_file_name = f'{logs_paths}/{self.assignment.course_code}-{self.assignment.year}{self.assignment.semester}-{self.assignment.assignment}_{self.student_id}_{time.time()}.json'
        try:
            with open(json_file_name, 'w', encoding='utf-8') as f:
                json.dump(json_data, f, ensure_ascii=False, indent=4, default=str)
        except:
            print(f"Can not write log file into {logs_paths}. Please save locally this output as a json file")
            print(json.dumps(json_data, indent=4, default=str))
        finally:
            return json_file_name


class AssignmentMarkingStudent:
    def __init__(
        self,
        assignment_id: str,
        student_id: int,
        all_marks_df: pd.DataFrame,
        grades_df: pd.DataFrame,
    ):
        self.assignment_id = assignment_id
        self.student_id = student_id
        self.all_marks_df = all_marks_df.copy()
        self.grades_df = grades_df.copy()
        self.assignment_mask = None
        self.student_mask = None
        self.student_marks_df: pd.DataFrame = None
        self.group_name: str = None
        self.score: float = 0
        self.grade: str = None
        self.student_name: str = None
        self.assignment_abbrev: str = None

    def reset_score_and_grade(self):
        self.score = 0
        self.grade = None

    def filter_marks(self):
        self.assignment_mask = self.all_marks_df[PgAssignmentsTable.assignment_id].astype('str') == str(self.assignment_id)
        self.student_mask = self.all_marks_df[PgStudentsTable.student_id] == int(self.student_id)
        self.student_marks_df = self.all_marks_df[self.assignment_mask & self.student_mask].copy()
        self.student_marks_df = self.student_marks_df.sort_values(by=[PgAssignmentCriteriasTable.criteria_desc], ascending=True).reset_index()
        if len(self.student_marks_df) > 0:
            self.student_name = self.student_marks_df.at[0, PgStudentsTable.student_name]
            self.group_name = self.student_marks_df.at[0, PgStudentsTable.group_name]
            self.assignment_abbrev = self.student_marks_df.at[0, PgAssignmentsTable.assignment_abbrev]

    def set_score_and_grade(self):
        self.filter_marks()
        if len(self.student_marks_df) > 0:
            self.reset_score_and_grade()
            self.student_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_mark] = self.student_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_score] * self.student_marks_df[PgAssignmentCriteriasTable.criteria_weight]
            self.score = (self.student_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_mark]).sum()
            self.grade = get_grade(self.grades_df, self.score)

    def display_student_marks(self):
        self.set_score_and_grade()
        print(f"########## {self.student_name} ({self.student_id}) - {self.group_name} ##########\n")

        if self.grade:
            print(f"\tMARK FOR {self.assignment_abbrev}: {self.score} ({self.grade})\n")

            for row_id in range(len(self.student_marks_df)):
                print("--------------------------------------------------\n")
                print(f"CRITERIA '{self.student_marks_df.at[row_id, PgAssignmentCriteriasTable.criteria_desc]}'\n")
                print(f"  -> {self.student_marks_df.at[row_id, PgAssignmentCriteriaStudentScoresTable.criteria_student_score]} ({get_grade(self.grades_df, self.student_marks_df.at[row_id, PgAssignmentCriteriaStudentScoresTable.criteria_student_score])})\n")
                print(f"  -> {self.student_marks_df.at[row_id, PgAssignmentCriteriaStudentScoresTable.criteria_student_score_comments]}")
        else:
            print("No grade found in database for this student")


class AssignmentSubcriteriaMarkingStudent:
    def __init__(
        self,
        assignment_id: str,
        student_id: int,
        all_marks_df: pd.DataFrame,
        grades_df: pd.DataFrame,
    ):
        self.assignment_id = assignment_id
        self.student_id = student_id
        self.all_marks_df = all_marks_df.copy()
        self.grades_df = grades_df.copy()
        self.assignment_mask = None
        self.student_mask = None
        self.student_marks_df: pd.DataFrame = None
        self.group_name: str = None
        self.score: float = 0
        self.grade: str = None
        self.student_name: str = None
        self.assignment_abbrev: str = None

    def reset_score_and_grade(self):
        self.score = 0
        self.grade = None

    def filter_marks(self):
        self.assignment_mask = self.all_marks_df[PgAssignmentsTable.assignment_id].astype('str') == str(self.assignment_id)
        self.student_mask = self.all_marks_df[PgStudentsTable.student_id] == int(self.student_id)
        self.student_marks_df = self.all_marks_df[self.assignment_mask & self.student_mask].copy()
        self.student_marks_df = self.student_marks_df.sort_values(by=[PgAssignmentSubcriteriasTable.subcriteria_desc], ascending=True).reset_index()
        if len(self.student_marks_df) > 0:
            self.student_name = self.student_marks_df.at[0, PgStudentsTable.student_name]
            self.group_name = self.student_marks_df.at[0, PgStudentsTable.group_name]
            self.assignment_abbrev = self.student_marks_df.at[0, PgAssignmentsTable.assignment_abbrev]


    def set_score_and_grade(self):
        self.filter_marks()
        if len(self.student_marks_df) > 0:
            self.reset_score_and_grade()
            self.student_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_mark] = self.student_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_score] * self.student_marks_df[PgAssignmentCriteriasTable.criteria_weight]
            self.score = (self.student_marks_df[PgAssignmentCriteriaStudentScoresTable.criteria_student_mark]).sum()
            self.grade = get_grade(self.grades_df, self.score)

    def display_student_marks(self):
        self.set_score_and_grade()
        print(f"########## {self.student_name} ({self.student_id}) - {self.group_name} ##########\n")

        if len(self.student_marks_df) > 0:
            print(f"\tMARK FOR {self.assignment_abbrev}: {self.score} ({self.grade})\n")

            crit_ids_list = self.student_marks_df[PgAssignmentCriteriasTable.criteria_id].unique()
            subcrit_ids_list = self.student_marks_df[PgAssignmentSubcriteriasTable.subcriteria_id].unique()

            for crit_id in crit_ids_list:
                crit_mask = self.student_marks_df[PgAssignmentCriteriasTable.criteria_id] == crit_id
                crit_desc = self.student_marks_df.loc[crit_mask, PgAssignmentCriteriasTable.criteria_desc].drop_duplicates()
                print("=============================================================")
                print("=")
                print(f"=  CRITERIA '{crit_desc[0]}'")
                print("=")
                print("=============================================================\n")
                
                for subcrit_id in subcrit_ids_list:
                    subcrit_mask = self.student_marks_df[PgAssignmentSubcriteriasTable.subcriteria_id] == subcrit_id
                    row_id = self.student_marks_df.index.get_loc(self.student_marks_df[subcrit_mask].index[0])
                    subcrit_desc = self.student_marks_df.at[row_id, PgAssignmentSubcriteriasTable.subcriteria_desc]

                    print(f"SUBCRITERIA '{subcrit_desc}'\n")
                    print(f"  -> {self.student_marks_df.at[row_id, PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score]} ({get_grade(self.grades_df, self.student_marks_df.at[row_id, PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score])})\n")
                    print(f"  -> {self.student_marks_df.at[row_id, PgAssignmentSubcriteriaStudentScoresTable.subcriteria_student_score_comments]}")
                    print("-------------------------------\n")
        else:
            print("No grade found in database for this student")



from dataclasses import dataclass

@dataclass
class PgGradesTable:
    name: str = "grades"
    grade_abbrev: str = 'grade_abbrev'
    grade_name: str = 'grade_name'
    grade_desc: str = 'grade_desc'
    grade_min: str = 'grade_min'
    grade_max: str = 'grade_max'


import os
import time
import json
from pathlib import Path
import pandas as pd
import altair as alt
from sqlalchemy import create_engine, MetaData, Table, and_, update
from sqlalchemy.sql import column
from pydantic import BaseModel


from utsam.bases.params import SemesterParams, _LOGS_, ZSCORE_COL
from utsam.bases.criterias import Criteria
from utsam.bases.tutors import AssignmentCriteriaMarkingTutor
from utsam.dbs.schemas.masterdata.students import PgStudentsTable
from utsam.dbs.schemas.masterdata.criterias import PgAssignmentCriteriasTable
from utsam.dbs.schemas.masterdata.employees import PgEmployeesTable
from utsam.dbs.schemas.masterdata.assignments import PgAssignmentsTable
from utsam.dbs.schemas.transactions.assignment_criteria_student_scores import PgAssignmentCriteriaStudentScoresTable


class AssignmentModel(BaseModel):
    course_code: int
    semester: int
    year: float
    assignment: str
    assignment_id: str

    @property
    def session(self):
        return f"{self.year}-{self.semester}" if self.semester else self.year


class Assignment:
    def __init__(
        self,
        course_code: int,
        semester: str,
        year: str,
        assignment: str,
        assignment_id: str,
    ):
        self.course_code: int = course_code
        self.semester: int = semester
        self.year: float = year
        self.assignment: str = assignment
        self.session = f"{self.year}-{self.semester}" if self.semester else self.year
        self.assignment_id = assignment_id


class AssignmentDir(Assignment):
    def __init__(
        self,
        course_code: int,
        semester: str,
        year: str,
        assignment: str,
        assignment_id: str,
    ):
        super().__init__(
            course_code = course_code,
            semester = semester,
            year = year,
            assignment = assignment,
            assignment_id = assignment_id,
        )
        self.at_path = None
        self.data_folder = 'data'
        self.data_raw_folder = '0-raw'
        self.data_int_folder = '1-interim'
        self.data_final_folder = '2-final'
        self.data_path = None
        self.raw_path = None
        self.interim_path = None
        self.final_path = None
        self.seed = (int(self.year) * 10) + 1 if self.semester == SemesterParams.autumn else 2

    def set_paths(self):
        self.at_path = Path(os.getcwd())
        self.data_path = self.at_path / self.data_folder
        self.raw_path = self.data_path / self.data_raw_folder
        self.interim_path = self.data_path / self.data_int_folder
        self.final_path = self.data_path / self.data_final_folder

        for item in [self.raw_path, self.interim_path, self.final_path]:
            item.mkdir(parents=True, exist_ok=True)

    #def set_colab(self):
    #    self.at_path = Path(f'content/MyDrive/{course}/{session}/Assignments/{assignment}')
    #    from google.colab import drive
    #
    #    drive.mount('./content')


class AssignmentMarking(AssignmentDir):
    def __init__(
        self,
        course_code: int,
        semester: str,
        year: str,
        assignment: str,
        assignment_id: str,
        criterias_df: pd.DataFrame,
        all_marks_df: pd.DataFrame,
    ):
        super().__init__(
            course_code = course_code,
            semester = semester,
            year = year,
            assignment = assignment,
            assignment_id = assignment_id,
        )
        self.criterias_df = criterias_df.copy()
        self.criteria_ids: dict = {}
        self.criteria_dict: dict = {}
        self.criteria_tutor_dict: dict = {}
        self.all_marks_df = all_marks_df.copy()
        self.tutor_names = list(self.all_marks_df[PgEmployeesTable.employee_first_name].unique())
        self.score = 0
        self.grade = None
        self.all_new_marks_df: pd.DataFrame = None
        self.new_score_col = f'new_{PgAssignmentCriteriaStudentScoresTable.criteria_student_score}'
        self.weighted_score_col = f'weighted_{PgAssignmentCriteriaStudentScoresTable.criteria_student_score}'
        self.new_weighted_score_col = f'new_{self.weighted_score_col}'
        self.set_paths()
        self.grouping_cols = [
            PgAssignmentsTable.assignment_id,
            PgStudentsTable.group_id,
            PgStudentsTable.student_id,
            PgStudentsTable.student_name,
            PgEmployeesTable.employee_first_name,
        ]
        self.student_marks: pd.DataFrame = None

    @property
    def n_criterias(self):
        return len(self.criteria_ids)
        
    def filter_at(self):
        criteria_mask = self.criterias_df[PgAssignmentsTable.assignment_id].astype('str') == str(self.assignment_id)
        self.criterias_df = self.criterias_df[criteria_mask]

    def set_at_criterias(self):
        self.filter_at()
        self.criteria_ids = {i: crit_id for i, crit_id in enumerate(self.criterias_df[PgAssignmentCriteriasTable.criteria_id].unique())}
        self.criteria_dict = {k: Criteria(criteria_id=crit_id) for k, crit_id in self.criteria_ids.items()}
        self.criteria_tutor_dict = {k: None for k, _ in self.criteria_ids.items()}

    def set_at_tutor_criterias(self):
        self.set_at_criterias()

        for i, crit_id in self.criteria_ids.items():
            tutor_crit_list = []

            for tutor_name in self.tutor_names:
                tutor_crit = AssignmentCriteriaMarkingTutor(
                    assignment_id=self.assignment_id,
                    criteria_id=crit_id,
                    tutor_first_name=tutor_name,
                    all_marks_df=self.all_marks_df,
                )
                tutor_crit.standardise(new_mean=75, new_std=15)

                tutor_crit_list.extend([tutor_crit])

            self.criteria_tutor_dict[i] = tutor_crit_list

    def get_all_marks(self):
        all_marks_list = []
        for _, crit in self.criteria_tutor_dict.items():
            for tutor in crit:
                all_marks_list.extend([tutor.tutor_marks_df])
        self.all_new_marks_df = pd.concat(all_marks_list, axis=0)


    def standardise_tutor_criteria(self, new_mean=85, new_std=5):
        self.set_at_tutor_criterias()
        for i, tutor_crit_marks in self.criteria_tutor_dict.items():
            for tutor_marks in tutor_crit_marks:
                tutor_marks.standardise(new_mean=new_mean, new_std=new_std)
        self.get_all_marks()

    def find_tutor_rank_by_name(self, tutor_name):
        return self.tutor_names.index(tutor_name)

    def plot_hist_by_criteria_tutor(self, criteria_rank, tutor_rank, marks_type='raw', employee_col_name='employee_first_name', bins=20):
        crit_df = self.criteria_tutor_dict.get(criteria_rank)[tutor_rank].tutor_marks_df.copy().reset_index(False)
        col_name = PgAssignmentCriteriaStudentScoresTable.criteria_student_score if marks_type == 'raw' else self.new_score_col
        crit_desc = crit_df[PgAssignmentCriteriasTable.criteria_desc].unique()[0]
        tutor_name = crit_df[PgEmployeesTable.employee_first_name].unique()[0]

        return alt.Chart(crit_df, title=f'Histogram of {marks_type} marks by {tutor_name} for {crit_desc}').mark_bar(
            opacity=0.3,
            binSpacing=0
        ).encode(
            alt.X(f'{col_name}:Q', bin = alt.BinParams(maxbins=bins)),
            alt.Y('count()'),
            alt.Color(f'{employee_col_name}:N')
        )
    
    def plot_criteria_hist_by_tutor(self, criteria_rank, marks_type='raw', employee_col_name='employee_first_name', bins=20):
        crit_df = self.all_marks_for_criteria(criteria_rank=criteria_rank)
        col_name = 'criteria_student_score' if marks_type == 'raw' else self.new_score_col
        crit_desc = crit_df[PgAssignmentCriteriasTable.criteria_desc].unique()[0]

        return alt.Chart(crit_df, title=f'Histogram of {marks_type} marks for {crit_desc}').mark_bar(
            opacity=0.3,
            binSpacing=0
        ).encode(
            alt.X(f'{col_name}:Q', bin = alt.BinParams(maxbins = 20)),
            alt.Y('count()', stack=None),
            facet=alt.Facet(f'{employee_col_name}:N', columns=1),
            tooltip=[
                f'{employee_col_name}:N', 
                alt.Tooltip(f'{col_name}:Q', bin = alt.BinParams(maxbins=bins)),
                'count()',
            ],
        ).resolve_axis(
            x='independent',
            y='independent',
        ).properties(
            width=600,
            height=250
        )

    def plot_at_marks_hist_by_tutor(self, marks_type='raw',employee_col_name='employee_first_name', bins=20):
        col_name = PgAssignmentCriteriaStudentScoresTable.criteria_student_score if marks_type == 'raw' else self.new_score_col

        return alt.Chart(self.student_marks, title=f'Histogram of {marks_type} marks by tutor').mark_bar(
            opacity=0.3,
            binSpacing=0
        ).encode(
            alt.X(f'{col_name}:Q', bin = alt.BinParams(maxbins=20)),
            alt.Y('count()'),
            facet=alt.Facet(f'{employee_col_name}:N', columns=1),
            tooltip=[
                f'{employee_col_name}:N', 
                alt.Tooltip(f'{col_name}:Q', bin = alt.BinParams(maxbins=bins)),
                'count()',
            ],
        ).resolve_axis(
            x='independent',
            y='independent',
        ).properties(
            width=600,
            height=250
        )
            
    def all_marks_for_criteria(self, criteria_rank):
        all_marks_list = []
        for tutor in self.criteria_tutor_dict.get(criteria_rank):
            all_marks_list.extend([tutor.tutor_marks_df])
        return pd.concat(all_marks_list, axis=0)


    def plot_hist_by_criteria(self, criteria_rank, marks_type='raw', employee_col_name='employee_first_name', bins=20):
        crit_df = self.all_marks_for_criteria(criteria_rank=criteria_rank)
        col_name = 'criteria_student_score' if marks_type == 'raw' else self.new_score_col

        return alt.Chart(crit_df, title=f'Histogram of {marks_type} marks').mark_bar(
            opacity=0.3,
            binSpacing=0
        ).encode(
            alt.X(f'{col_name}:Q', bin = alt.BinParams(maxbins=bins)),
            alt.Y('count()'),
            alt.Color(f'{employee_col_name}:N'),
            column=f'{PgAssignmentCriteriasTable.criteria_desc}:N'
        )

    def plot_hists(self, marks_type='raw', employee_col_name=PgEmployeesTable.employee_first_name, bins=20):
        col_name = PgAssignmentCriteriaStudentScoresTable.criteria_student_score if marks_type == 'raw' else self.new_score_col

        return alt.Chart(self.all_new_marks_df, title=f'Histogram of {marks_type} marks').mark_bar(
            opacity=0.3,
            binSpacing=0
        ).encode(
            alt.X(f'{col_name}:Q', bin = alt.BinParams(maxbins = 20)),
            alt.Y('count()', stack=None),
            alt.Color(f'{employee_col_name}:N'),
            facet=alt.Facet(f'{PgAssignmentCriteriasTable.criteria_desc}:N', columns=1),
            tooltip=[
                f'{employee_col_name}:N', 
                alt.Tooltip(f'{col_name}:Q', bin = alt.BinParams(maxbins=bins)),
                'count()',
            ],
        ).resolve_axis(
            x='independent',
            y='independent',
        ).properties(
            width=600,
            height=250
        )

    def reset_score_and_grade(self):
        self.score = 0
        self.grade = None

    def set_score_and_grade(self):
        if len(self.all_new_marks_df) > 0:
            self.reset_score_and_grade()

            result_cols = [PgAssignmentCriteriaStudentScoresTable.criteria_student_score, self.new_score_col]
            temp_df = self.all_new_marks_df.copy()

            for col_name in result_cols:
                temp_df[col_name] = temp_df[col_name] * temp_df[PgAssignmentCriteriasTable.criteria_weight]

            self.student_marks = temp_df.groupby(self.grouping_cols).agg({col_name:'sum' for col_name in result_cols}).reset_index().sort_values([PgStudentsTable.student_name])

    def standardise_criteria_by_tutor(self, tutor_name, new_mean=85, new_std=5):
        # Todo: standardise with specifc mean and std for a single tutor
        pass

    def to_dict(self):
        if len(self.criteria_dict) > 0:
            return {
                "student_id": self.student_id,
                "student_name": self.student_name,
                "tutor_name": self.tutor_name,
                "course_code": self.assignment.course_code,
                "semester": self.assignment.semester,
                "year": self.assignment.year,
                "assignment": self.assignment.assignment,
                "grade": self.grade,
                "score": self.score,
                #"comments": self.comments,
                "criterias": {crit_id: crit.to_dict() for crit_id, crit in self.criteria_dict.items()}
            }

    def to_json(self):
        logs_paths = Path.cwd() / _LOGS_ / self.tutor_name
        logs_paths.mkdir(parents=True, exist_ok=True)
        json_data = self.to_dict()
        json_file_name = f'{logs_paths}/{self.assignment.course_code}-{self.assignment.year}{self.assignment.semester}-{self.assignment.assignment}_{self.student_id}_{time.time()}.json'
        try:
            with open(json_file_name, 'w', encoding='utf-8') as f:
                json.dump(json_data, f, ensure_ascii=False, indent=4)
        except:
            print(f"Can not write log file into {logs_paths}. Please save locally this output as a json file")
            print(json.dumps(json_data, indent=4))
        finally:
            return json_file_name

    def to_sql(self):
        json_data = self.to_dict()

        update_stmt = (
            update(user_table)
            .where(user_table.c.id == address_table.c.user_id)
            .where(address_table.c.email_address == "patrick@aol.com")
            .values(
                {
                    user_table.c.fullname: "Pat",
                    address_table.c.email_address: "pat@aol.com",
                }
            )
        )
        print(update_stmt)



from typing import Optional
from pydantic import BaseModel


class Student(BaseModel):
    student_id: int
    canvas_student_id: int
    student_name: str
    student_email: str
    student_status: str

    def __hash__(self):
        return hash((self.student_id))



from typing import Optional
from pydantic import BaseModel


class Employee(BaseModel):
    employee_id: str
    employee_first_name: str
    employee_last_name: str
    staff_id: int

    def __hash__(self):
        return hash((self.staff_id))


from utsam.dbs.schemas.transactions.assignment_criteria_student_scores import PgAssignmentCriteriaStudentScoresTable
from utsam.dbs.schemas.transactions.assignment_subcriteria_grade import PgAssignmentSubcriteriaGrade
from utsam.dbs.schemas.transactions.assignment_subcriteria_student_scores import PgAssignmentSubcriteriaStudentScoresTable
from utsam.dbs.schemas.transactions.assignment_student_scores import PgAssignmentStudentScoresTable


from dataclasses import dataclass

@dataclass
class PgAssignmentCriteriaStudentScoresTable:
    name: str = "assignment_criteria_student_scores"
    criteria_student_score_id: str = 'criteria_student_score_id'
    criteria_student_score: str = 'criteria_student_score'
    criteria_student_score_comments: str = 'criteria_student_score_comments'
    criteria_student_mark: str = 'criteria_student_mark'
    criteria_grade_abbrev: str = 'criteria_grade_abbrev'
    criteria_grade_desc: str = 'criteria_grade_desc'




from typing import Optional
from pydantic import BaseModel


class Semester(BaseModel):
    semester_id: str
    semester_name: str
    semester_abbrev2: str
    semester_abbrev3: str

    def __hash__(self):
        return hash((self.semester_name))


from utsam.dbs.schemas.masterdata.grades import PgGradesTable
from utsam.bases.params import _SELECT_

def get_score(grades_df, grade, finetuning, na_value=_SELECT_):
    if grade == na_value:
        return 0
    temp_df = grades_df.copy()
    temp_df.set_index(PgGradesTable.grade_abbrev, inplace=True)
    grade_min = temp_df.at[grade, PgGradesTable.grade_min]
    grade_max = temp_df.at[grade, PgGradesTable.grade_max]
    return grade_min + (grade_max - grade_min) * finetuning / 100


def get_grade(grades_df, score):
    temp_df = grades_df.copy()
    min_grade_mask = temp_df[PgGradesTable.grade_min] <= score
    max_grade_mask = temp_df[PgGradesTable.grade_max] > score
    
    grade = temp_df.loc[min_grade_mask & max_grade_mask, PgGradesTable.grade_abbrev].values
    if len(grade) > 0:
        return grade[0]
    return None